from typing import Optional

from dolomite_base.read_object import read_object_registry
from genomicranges import CompressedGenomicRangesList

from .read_atomic_vector_list import _read_compressed_list

read_object_registry["genomic_ranges_list"] = "dolomite_ranges.read_genomic_ranges_list"


def read_genomic_ranges_list(path: str, metadata: Optional[dict], **kwargs) -> CompressedGenomicRangesList:
    """Load genomic ranges into a
    :py:class:`~genomicranges.grangeslist.CompressedGenomicRangesList` object.

    This method
    should generally not be called directly but instead be invoked by
    :py:meth:`~dolomite_base.read_object.read_object`.

    Args:
        path:
            Path to the directory containing the object.

        metadata:
            Metadata for the object.

        kwargs:
            Further arguments, ignored.

    Returns:
        A :py:class:`~genomicranges.grangeslist.CompressedGenomicRangesList` object.
    """

    return _read_compressed_list(path, metadata=metadata, name="genomic_ranges_list", **kwargs)



from typing import Optional

from compressed_lists import CompressedSplitBiocFrameList
from dolomite_base.read_object import read_object_registry

from .read_atomic_vector_list import _read_compressed_list

read_object_registry["data_frame_list"] = "dolomite_ranges.read_data_frame_list"


def read_data_frame_list(path: str, metadata: Optional[dict], **kwargs) -> CompressedSplitBiocFrameList:
    """Load data frame list.

    This method
    should generally not be called directly but instead be invoked by
    :py:meth:`~dolomite_base.read_object.read_object`.

    Args:
        path:
            Path to the directory containing the object.

        metadata:
            Metadata for the object.

        kwargs:
            Further arguments, ignored.

    Returns:
        A :py:class:`~compressed_lists.biocframe_list.CompressedSplitBiocFrameList` object.
    """

    return _read_compressed_list(path, metadata=metadata, name="data_frame_list", **kwargs)



import os
from typing import Optional

import dolomite_base as dl
import h5py
from dolomite_base.read_object import read_object_registry
from genomicranges import SeqInfo

read_object_registry[
    "sequence_information"
] = "dolomite_ranges.read_sequence_information"


def read_sequence_information(path: str, metadata: Optional[dict], **kwargs) -> SeqInfo:
    """Load sequence information into a
    :py:class:`~genomicranges.SeqInfo.SeqInfo` object.

    This method should generally not be called directly but instead be
    invoked by :py:meth:`~dolomite_base.read_object.read_object`.

    Args:
        path:
            Path to the directory containing the object.

        metadata:
            Metadata for the object.

        kwargs:
            Further arguments, ignored.

    Returns:
        A :py:class:`~genomicranges.SeqInfo.SeqInfo` object.
    """

    with h5py.File(os.path.join(path, "info.h5"), "r") as handle:
        ghandle = handle["sequence_information"]

        seqnames = dl.load_vector_from_hdf5(
            ghandle["name"], expected_type=str, report_1darray=True
        )

        seqlengths = dl.load_vector_from_hdf5(
            ghandle["length"], expected_type=int, report_1darray=True
        )

        is_circular = dl.load_vector_from_hdf5(
            ghandle["circular"], expected_type=bool, report_1darray=True
        )

        genome = dl.load_vector_from_hdf5(
            ghandle["genome"], expected_type=str, report_1darray=True
        )

    return SeqInfo(seqnames, seqlengths, is_circular, genome)



import os
from typing import Optional

import dolomite_base as dl
import h5py
from biocutils import combine_sequences
from compressed_lists import CompressedList


@dl.save_object.register
@dl.validate_saves
def save_atomic_vector_list(x: CompressedList, path: str, **kwargs):
    """Method for saving :py:class:`~CompressedList`
    objects to their corresponding file representations, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x:
            Object to be staged.

        path:
            Path to a directory in which to save ``x``.

        data_frame_args:
            Further arguments to pass to the ``save_object`` method for
            ``mcols``.

        kwargs:
            Further arguments to be passed to individual methods.

    Returns:
        `x` is saved to `path`.
    """
    return _save_compressed_list(x, path=path, name="atomic_vector_list", **kwargs)


def _save_compressed_list(x, path, name, **kwargs):
    os.mkdir(path)

    _all_ranges = x.get_unlist_data()
    if isinstance(_all_ranges, list) and len(_all_ranges) > 1:
        _all_ranges = combine_sequences(*x.get_unlist_data())

    dl.alt_save_object(_all_ranges, path=os.path.join(path, "concatenated"), **kwargs)

    _elem_annotation = x.get_element_metadata()
    if _elem_annotation is not None:
        dl.alt_save_object(_elem_annotation, path=os.path.join(path, "element_annotations"), **kwargs)

    _meta = x.get_metadata()
    if _meta is not None and len(_meta) > 0:
        dl.alt_save_object(_meta, path=os.path.join(path, "other_annotations"), **kwargs)

    with h5py.File(os.path.join(path, "partitions.h5"), "w") as handle:
        ghandle = handle.create_group(name)

        dl.write_integer_vector_to_hdf5(ghandle, name="lengths", h5type="u4", x=x.get_element_lengths())

        if x.get_names() is not None:
            dl.write_string_vector_to_hdf5(ghandle, name="names", x=x.get_names())

    _info = {}
    _info[name] = {"version": "1.0"}
    dl.save_object_file(path, name, _info)

    return



import os
from tempfile import mkdtemp

from biocframe import BiocFrame
from dolomite_base import read_object, save_object
from genomicranges import GenomicRanges, CompressedGenomicRangesList
from iranges import IRanges
import numpy as np
import dolomite_ranges


def test_genomic_ranges_list():
    a = GenomicRanges(
        seqnames=["chr1", "chr2", "chr1", "chr3"],
        ranges=IRanges([1, 3, 2, 4], [10, 30, 50, 60]),
        strand=["-", "+", "*", "+"],
        mcols=BiocFrame({"score": [1, 2, 3, 4]}),
    )

    b = GenomicRanges(
        seqnames=["chr2", "chr4", "chr5"],
        ranges=IRanges([3, 6, 4], [30, 50, 60]),
        strand=["-", "+", "*"],
        mcols=BiocFrame({"score": [2, 3, 4]}),
    )

    grl = CompressedGenomicRangesList.from_list(lst=[a, b], names=["a", "b"])

    dir = os.path.join(mkdtemp(), "granges")
    save_object(grl, dir)

    roundtrip = read_object(dir)
    assert isinstance(roundtrip, CompressedGenomicRangesList)
    assert roundtrip.get_names() == grl.get_names()
    assert len(roundtrip.get_unlist_data()) == len(grl.get_unlist_data())
    assert np.allclose(roundtrip["a"].start, grl["a"].start)
    assert np.allclose(roundtrip["a"].strand, grl["a"].strand)


def test_genomic_ranges_list_empty():
    grl = CompressedGenomicRangesList.empty(n=100)

    dir = os.path.join(mkdtemp(), "granges_empty")
    save_object(grl, dir)

    roundtrip = read_object(dir)
    assert roundtrip.get_names() == grl.get_names()
    assert len(roundtrip.get_unlist_data()) == len(grl.get_unlist_data())
    assert np.allclose(roundtrip.get_element_lengths(), grl.get_element_lengths())


import os
from typing import Optional

import dolomite_base as dl
import h5py
from compressed_lists import Partitioning, splitAsCompressedList
from dolomite_base.read_object import read_object_registry

read_object_registry["atomic_vector_list"] = "dolomite_ranges.read_atomic_vector_list"


def read_atomic_vector_list(path: str, metadata: Optional[dict], **kwargs):
    """Load a list of atomic vectors from its on-disk representation.

    Args:
        path:
            Path to the directory containing the object.

        metadata:
            Metadata for the object.

        kwargs:
            Further arguments, ignored.


    """
    return _read_compressed_list(path, metadata, "atomic_vector_list", **kwargs)


def _read_compressed_list(path: str, metadata: Optional[dict], name: str, **kwargs):
    concat_path = os.path.join(path, "concatenated")
    concat = dl.alt_read_object(concat_path, **kwargs)

    fpath = os.path.join(path, "partitions.h5")
    with h5py.File(fpath, "r") as fhandle:
        ghandle = fhandle[name]
        lengths = dl.load_vector_from_hdf5(ghandle["lengths"], expected_type=int, report_1darray=True)

        names = None
        if "names" in ghandle:
            names = dl.load_vector_from_hdf5(ghandle["names"], expected_type=str, report_1darray=True)

        output = splitAsCompressedList(concat, Partitioning.from_lengths(lengths=lengths, names=names))

    _elem_annotation_path = os.path.join(path, "element_annotations")
    if os.path.exists(_elem_annotation_path):
        _mcols = dl.alt_read_object(_elem_annotation_path, **kwargs)
        output = output.set_element_metadata(_mcols)

    _meta_path = os.path.join(path, "other_annotations")
    if os.path.exists(_meta_path):
        _meta = dl.alt_read_object(_meta_path, **kwargs)
        output = output.set_metadata(_meta.as_dict())

    return output



import os
from typing import Optional

import dolomite_base as dl
import h5py
from dolomite_base.read_object import read_object_registry
from genomicranges import GenomicRanges
from iranges import IRanges

read_object_registry["genomic_ranges"] = "dolomite_ranges.read_genomic_ranges"


def read_genomic_ranges(path: str, metadata: Optional[dict], **kwargs) -> GenomicRanges:
    """Load genomic ranges into a
    :py:class:`~genomicranges.GenomicRanges.GenomicRanges` object.

    This method
    should generally not be called directly but instead be invoked by
    :py:meth:`~dolomite_base.read_object.read_object`.

    Args:
        path:
            Path to the directory containing the object.

        metadata:
            Metadata for the object.

        kwargs:
            Further arguments, ignored.

    Returns:
        A :py:class:`~genomicranges.GenomicRanges.GenomicRanges` object.
    """
    _seqinfo_path = os.path.join(path, "sequence_information")
    seqinfo = dl.alt_read_object(path=_seqinfo_path, **kwargs)

    with h5py.File(os.path.join(path, "ranges.h5"), "r") as handle:
        ghandle = handle["genomic_ranges"]

        seqnames = dl.load_vector_from_hdf5(
            ghandle["sequence"], expected_type=int, report_1darray=True
        )

        starts = dl.load_vector_from_hdf5(
            ghandle["start"], expected_type=int, report_1darray=True
        )

        widths = dl.load_vector_from_hdf5(
            ghandle["width"], expected_type=int, report_1darray=True
        )

        strand = dl.load_vector_from_hdf5(
            ghandle["strand"], expected_type=int, report_1darray=True
        )

    gr = GenomicRanges(
        seqnames=seqnames,
        ranges=IRanges(starts, widths),
        strand=strand,
        seqinfo=seqinfo,
    )

    _range_annotation_path = os.path.join(path, "range_annotations")
    if os.path.exists(_range_annotation_path):
        _mcols = dl.alt_read_object(_range_annotation_path, **kwargs)
        gr = gr.set_mcols(_mcols)

    _meta_path = os.path.join(path, "other_annotations")
    if os.path.exists(_meta_path):
        _meta = dl.alt_read_object(_meta_path, **kwargs)
        gr = gr.set_metadata(_meta.as_dict())

    return gr



import os
from typing import Optional

import dolomite_base as dl
import h5py
from genomicranges import GenomicRanges


@dl.save_object.register
@dl.validate_saves
def save_genomic_ranges(
    x: GenomicRanges, path: str, data_frame_args: Optional[dict] = None, **kwargs
):
    """Method for saving :py:class:`~genomicranges.GenomicRanges.GenomicRanges`
    objects to their corresponding file representations, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x:
            Object to be staged.

        path:
            Path to a directory in which to save ``x``.

        data_frame_args:
            Further arguments to pass to the ``save_object`` method for
            ``mcols``.

        kwargs:
            Further arguments to be passed to individual methods.

    Returns:
        `x` is saved to `path`.
    """
    os.mkdir(path)

    if data_frame_args is None:
        data_frame_args = {}

    _info = {"genomic_ranges": {"version": "1.0"}}
    dl.save_object_file(path, "genomic_ranges", _info)

    # sequence information
    spath = os.path.join(path, "sequence_information")
    dl.save_object(x.get_seqinfo(), spath)

    with h5py.File(os.path.join(path, "ranges.h5"), "w") as handle:
        ghandle = handle.create_group("genomic_ranges")

        _seqnames = x.get_seqnames(as_type="factor")
        dl.write_integer_vector_to_hdf5(
            ghandle, name="sequence", h5type="u4", x=_seqnames.get_codes()
        )

        _ranges = x.get_ranges()
        dl.write_integer_vector_to_hdf5(
            ghandle, name="start", h5type="i4", x=_ranges.get_start()
        )
        dl.write_integer_vector_to_hdf5(
            ghandle, name="width", h5type="u4", x=_ranges.get_width()
        )

        dl.write_integer_vector_to_hdf5(
            ghandle, name="strand", h5type="i4", x=x.get_strand()
        )

        if x.get_names() is not None:
            dl.write_string_vector_to_hdf5(ghandle, name="name", x=x.get_names())

    _range_annotation = x.get_mcols()
    if _range_annotation is not None and _range_annotation.shape[1] > 0:
        dl.alt_save_object(
            _range_annotation,
            path=os.path.join(path, "range_annotations"),
            **data_frame_args,
        )

    _meta = x.get_metadata()
    if _meta is not None and len(_meta) > 0:
        dl.alt_save_object(
            _meta, path=os.path.join(path, "other_annotations"), **kwargs
        )

    return



import os
from tempfile import mkdtemp

import dolomite_base as dl
import dolomite_ranges
from genomicranges import SeqInfo

def test_sequence_information():
    si = SeqInfo(
        seqnames=["chrA", "chrB", "chrC"],
        seqlengths=[10, 100, 2200],
        is_circular=[False, True, False],
        genome=["hg19", "hg38", "hg19"],
    )

    dir = os.path.join(mkdtemp(), "seqinfo")
    dl.save_object(si, dir)

    roundtrip = dl.read_object(dir)
    assert roundtrip.get_seqnames() == si.get_seqnames()
    assert roundtrip.get_seqlengths() == si.get_seqlengths()
    assert roundtrip.get_is_circular() == si.get_is_circular()
    assert roundtrip.get_genome() == si.get_genome()


def test_sequence_information_with_none():
    si = SeqInfo(
        seqnames=["chrA", "chrB", "chrC"],
        seqlengths=[10, None, 2200],
        is_circular=[None, True, False],
        genome=["hg19", "hg38", None],
    )

    dir = os.path.join(mkdtemp(), "seqinfo")
    dl.save_object(si, dir)

    roundtrip = dl.read_object(dir)
    assert roundtrip.get_seqnames() == si.get_seqnames()
    assert roundtrip.get_seqlengths() == si.get_seqlengths()
    assert roundtrip.get_is_circular() == si.get_is_circular()
    assert roundtrip.get_genome() == si.get_genome()


def test_sequence_information_all_none():
    si = SeqInfo(
        ["chrA", "chrB", "chrC"],
        [None] * 3,
        [None] * 3,
        [None] * 3
    )

    dir = os.path.join(mkdtemp(), "seqinfo")
    dl.save_object(si, dir)

    roundtrip = dl.read_object(dir)
    assert roundtrip.get_seqnames() == si.get_seqnames()
    assert roundtrip.get_seqlengths() == si.get_seqlengths()
    assert roundtrip.get_is_circular() == si.get_is_circular()
    assert roundtrip.get_genome() == si.get_genome()



import os
from tempfile import mkdtemp

from dolomite_base import read_object, save_object
from genomicranges import GenomicRanges, SeqInfo
from iranges import IRanges
import dolomite_ranges


def test_genomic_ranges():
    gr = GenomicRanges(
        seqnames=["chrA", "chrB", "chrC"],
        ranges=IRanges([10, 30, 2200], [20, 50, 30]),
        strand=["*", "+", "-"],
    )

    dir = os.path.join(mkdtemp(), "granges")
    save_object(gr, dir)

    roundtrip = read_object(dir)
    assert roundtrip.get_seqnames() == gr.get_seqnames()
    assert (roundtrip.get_start() == gr.get_start()).all()
    assert (roundtrip.get_end() == gr.get_end()).all()
    assert (roundtrip.get_strand() == gr.get_strand()).all()
    assert roundtrip.get_seqinfo().get_seqlengths() == gr.get_seqinfo().get_seqlengths()


def test_genomic_ranges_full_load():
    gr = GenomicRanges(
        seqnames=["chrA", "chrB", "chrC"],
        ranges=IRanges([10, 30, 2200], [20, 50, 30]),
        strand=["*", "+", "-"],
    )

    gr.metadata = {"ARG": [5, 3, 2, 1]}
    gr.seq_info = SeqInfo(
        seqnames=["chrA", "chrB", "chrC"],
        seqlengths=[1000, 2000, 3000],
        is_circular=[False] * 3,
        genome=["hg38"] * 3,
    )

    dir = os.path.join(mkdtemp(), "granges2")
    save_object(gr, dir)

    roundtrip = read_object(dir)
    assert roundtrip.seqinfo.seqlengths == gr.seqinfo.seqlengths
    assert list(roundtrip.metadata.keys()) == list(gr.metadata.keys())



import dolomite_base as dl
from compressed_lists import CompressedSplitBiocFrameList

from .save_atomic_vector_list import _save_compressed_list


@dl.save_object.register
@dl.validate_saves
def save_compressed_genomic_ranges_list(x: CompressedSplitBiocFrameList, path: str, **kwargs):
    """Method for saving :py:class:`~compressed_lists.biocframe_list.CompressedSplitBiocFrameList`
    objects to their corresponding file representations, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x:
            Object to be staged.

        path:
            Path to a directory in which to save ``x``.

        data_frame_args:
            Further arguments to pass to the ``save_object`` method for
            ``mcols``.

        kwargs:
            Further arguments to be passed to individual methods.

    Returns:
        `x` is saved to `path`.
    """
    return _save_compressed_list(x, path=path, name="genomic_ranges_list", **kwargs)



import os

import dolomite_base as dl
import h5py
from genomicranges import SeqInfo


@dl.save_object.register
@dl.validate_saves
def save_sequence_information(x: SeqInfo, path: str, **kwargs):
    """Save Sequence information to disk.

    Args:
        x:
            Object to be saved.

        path:
            Path to a directory in which to save ``x``.

        kwargs:
            Further arguments to be passed to individual methods.

    Returns:
        `x` is saved to `path`.
    """
    os.mkdir(path)

    _info = {"sequence_information": {"version": "1.0"}}
    dl.save_object_file(path, "sequence_information", _info)

    with h5py.File(os.path.join(path, "info.h5"), "w") as handle:
        ghandle = handle.create_group("sequence_information")

        dl.write_string_vector_to_hdf5(ghandle, name="name", x=x.get_seqnames())
        dl.write_integer_vector_to_hdf5(
            ghandle, name="length", x=x.get_seqlengths(), h5type="u4"
        )

        dl.write_boolean_vector_to_hdf5(ghandle, name="circular", x=x.get_is_circular())
        dl.write_string_vector_to_hdf5(ghandle, name="genome", x=x.get_genome())

    return
